use crate::config::*;
use crate::utils::EpochTimestamp;
use actix_web::{
    error::ErrorInternalServerError, web::Data, Error as WebError, HttpResponse,
    Result as WebResult,
};
use reqwest::{Client, Url, UrlError};

use serde::Serialize;
use std::sync::{Arc, Mutex, MutexGuard};
use tera::{Context, Tera};

#[derive(Clone, Serialize, Default)]
pub struct Status {
    status: u8,
    location: String,
    domain: String,
    endpoint: String,
    error: Option<String>,
}

#[derive(Serialize)]
pub struct QueryResults {
    pub last_update: EpochTimestamp,
    pub refresh_time: u64,
    pub config: Config,
    pub statuses: Vec<Status>,
}

pub struct State(Arc<Mutex<QueryResults>>);

impl State {
    pub fn new(config: &Config) -> Self {
        State(Arc::from(Mutex::from(QueryResults {
            last_update: EpochTimestamp::now(),
            refresh_time: config.refresh_time.clone(),
            config: config.clone(),
            statuses: vec![],
        })))
    }

    fn lock(&self) -> MutexGuard<QueryResults> {
        self.0.lock().unwrap()
    }
}

pub fn index(tmpl: Data<Tera>, state: Data<State>) -> WebResult<HttpResponse, WebError> {
    let state = update_state(state.lock());
    let mut ctx = Context::new();
    ctx.insert("results", &*state);
    let s = tmpl.render("index.html", &ctx).map_err(|e| {
        println!("{:?}", e);
        ErrorInternalServerError("Template error")
    })?;
    Ok(HttpResponse::Ok().content_type("text/html").body(s))
}

pub fn json_endpoint(state: Data<State>) -> HttpResponse {
    let state = update_state(state.lock());
    HttpResponse::Ok().json(&state.statuses)
}

fn update_state(mut state: MutexGuard<QueryResults>) -> MutexGuard<QueryResults> {
    if EpochTimestamp::now() - state.last_update >= state.refresh_time {
        state.last_update = EpochTimestamp::now();
        state.statuses = update_status(&state.config);
    }

    state
}

fn update_status(config: &Config) -> Vec<Status> {
    let client = Client::new();
    let mut results: Vec<Status> = vec![];

    for website_conf in &config.websites {
        for endpoint in &website_conf.endpoints {
            results.push(get_result(website_conf, &client, endpoint));
        }
    }

    results
}

fn get_result(website_conf: &WebsiteConfig, client: &Client, endpoint: &EndpointConfig) -> Status {
    let (label, path, port, code, body) = get_endpoint_info(endpoint.clone());
    let url = get_url(&website_conf.base, &path, port).expect("reading config");
    let ping_result = client.get(&url).send();

    match ping_result {
        Ok(mut res) => {
            let res_body = res.text().expect("could not get body of request");
            let does_code_match = res.status() == code;
            let does_body_match = body.is_empty() || res_body == body;
            let mut error = None;

            if !does_code_match {
                error = Some(format!(
                    "Status code mismatch: {} != {}.",
                    res.status().as_u16(),
                    code
                ));
            }

            if !does_body_match {
                error = Some(if let Some(msg) = error {
                    format!(
                        "{} Body mismatch: {} != {}.",
                        msg,
                        res_body.len(),
                        body.len()
                    )
                } else {
                    format!("Body mismatch: {} != {}.", res_body.len(), body.len())
                });
            }

            Status {
                status: if error.is_some() { 1 } else { 0 },
                location: url,
                domain: website_conf.label.clone(),
                endpoint: label,
                error,
            }
        }
        Err(e) => Status {
            status: 2,
            location: url,
            domain: website_conf.label.clone(),
            endpoint: label,
            error: Some(format!("{}", e)),
        },
    }
}

fn get_url(base: &String, path: &String, port: Option<u16>) -> Result<String, UrlError> {
    let mut url = Url::parse(base)?.join(path)?;
    if let Err(e) = url.set_port(port) {
        println!("{:?}", e);
    }
    Ok(url.into_string())
}
